#
# This module accepts the following environment variables:
#
#    NETCDF_DIR or NETCDF_ROOT - Specify the location of NetCDF
#
# This module defines the following CMake variables:
#
#    NETCDF_FOUND - True if libnetcdf is found
#    NETCDF_Fortran_LIBRARY
#    NETCDF_Fortran_INCLUDE_DIR

# Copied and modified (nc-config -> nf-config) from FindNETCDF of VTK, which is:
# Distributed under the bsd license (the "License")
#
# This software is distributed WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See COPYING-CMAKE-SCRIPTS for more information.
#=============================================================================


if (UNIX AND NOT NETCDF_FOUND)
 # Use nf-config to obtain the libraries
 find_program (NETCDF_CONFIG nf-config
  HINTS
  ${NETCDF_DIR}
  ${NETCDF_ROOT}
  $ENV{NETCDF_DIR}
  $ENV{NETCDF_ROOT}
  PATHS
  PATH_SUFFIXES bin
  /sw 
  /opt/local 
  /opt
  /usr/local
 )

 if (NETCDF_CONFIG)
  execute_process (COMMAND ${NETCDF_CONFIG} --fflags
   ERROR_QUIET OUTPUT_STRIP_TRAILING_WHITESPACE
   OUTPUT_VARIABLE NETCDF_CONFIG_FFLAGS)
  if (NETCDF_CONFIG_FFLAGS)
   string (REGEX MATCHALL "(^| )-I[^ ]+" _netcdf_dashI ${NETCDF_CONFIG_FFLAGS})
   string (REGEX REPLACE "(^| )-I" "" _netcdf_includepath "${_netcdf_dashI}")
   string (REGEX REPLACE "(^| )-I[^ ]+" "" _netcdf_cflags_other ${NETCDF_CONFIG_FFLAGS})
  endif (NETCDF_CONFIG_FFLAGS)
  execute_process (COMMAND ${NETCDF_CONFIG} --flibs
   ERROR_QUIET OUTPUT_STRIP_TRAILING_WHITESPACE
   OUTPUT_VARIABLE NETCDF_CONFIG_LIBS)
  if (NETCDF_CONFIG_LIBS)
   # Ensure -l is precedeced by whitespace to not match
   # '-l' in '-L/usr/lib/x86_64-linux-gnu/hdf5/serial'
   string (REGEX MATCHALL "(^| )-l[^ ]+" _netcdf_dashl ${NETCDF_CONFIG_LIBS})
   string (REGEX REPLACE "(^| )-l" "" _netcdf_lib "${_netcdf_dashl}")
   string (REGEX MATCHALL "(^| )-L[^ ]+" _netcdf_dashL ${NETCDF_CONFIG_LIBS})
   string (REGEX REPLACE "(^| )-L" "" _netcdf_libpath "${_netcdf_dashL}")
  endif (NETCDF_CONFIG_LIBS)
        endif (NETCDF_CONFIG)
 if (_netcdf_lib)
  list (REMOVE_DUPLICATES _netcdf_lib)
 endif (_netcdf_lib)
endif (UNIX AND NOT NETCDF_FOUND)

find_path (NETCDF_Fortran_INCLUDE_DIR netcdf.inc
 HINTS
 ${_netcdf_includepath}
 ${NETCDF_DIR}
 ${NETCDF_ROOT}
 $ENV{NETCDF_DIR}
 $ENV{NETCDF_ROOT}
 PATH_SUFFIXES
 include/netcdf
 include/netcdf-4
 include/netcdf-3
 include
 PATHS
 /opt/local 
 /opt
 /usr/local
)

      
# find all libs that nf-config reports
foreach (_extralib ${_netcdf_lib})
        foreach (_extrapath ${_netcdf_libpath})
                find_library (_found_lib_${_extralib}
                ${_extralib}
                PATH ${_extrapath}
                NO_DEFAULT_PATH NO_PACKAGE_ROOT_PATH NO_CMAKE_PATH
                NO_CMAKE_ENVIRONMENT_PATH NO_SYSTEM_ENVIRONMENT_PATH
                NO_CMAKE_SYSTEM_PATH NO_CMAKE_FIND_ROOT_PATH)
        endforeach (_extrapath) 
        find_library (_found_lib_${_extralib} NAMES ${_extralib})
        list (APPEND NETCDF_Fortran_LIBRARY ${_found_lib_${_extralib}})
endforeach (_extralib)


include (FindPackageHandleStandardArgs)
find_package_handle_standard_args (NETCDF NETCDF_Fortran_LIBRARY NETCDF_Fortran_INCLUDE_DIR)



