
! Part of TsunAWI
! Copyright (c) 2014
! Alfred Wegener Institut,
! Helmholtz Centre for Polar and Marine Research
! Contact: rakowsky@awi.de


! MODULE: ELEMENTS -- contains the local element definitions, basis functions

MODULE ELEMENTS

    USE PARAMETERS
    IMPLICIT NONE
    SAVE 

    REAL(KIND=wp), ALLOCATABLE, DIMENSION(:,:)   :: bafu_xy
    REAL(KIND=wp), ALLOCATABLE, DIMENSION(:)     :: voltriangle
  
  END MODULE ELEMENTS

SUBROUTINE basisfunctions

  USE MESH
  USE ELEMENTS
    IMPLICIT NONE
  
    INTEGER                          :: el, i
    INTEGER                          :: singular_el=0

    REAL(KIND=dp), parameter         :: std_vol=0.5_dp

    REAL(KIND=dp), DIMENSION(2,2)    :: jacobian
    REAL(KIND=dp)                    :: DET
    REAL(kind=dp)                    :: DET_inv
    REAL(kind=dp), parameter         :: one_third = 1._dp/3._dp      


    ALLOCATE(bafu_xy(6,  myDim_elem2D + eDim_elem2D + exDim_elem2D))
    ALLOCATE(voltriangle(myDim_elem2D + eDim_elem2D + exDim_elem2D))
  
! bafu_xy = bafuzeta_xy, we only compute bafuzeta and 
! Employ: bafuvel_xy(1,el) = -2.*bafuzeta_xy(3,el)
!         bafuvel_xy(2,el) = -2.*bafuzeta_xy(1,el)
!         bafuvel_xy(3,el) = -2.*bafuzeta_xy(2,el)

!$OMP PARALLEL default(shared) private(el, i, DET, DET_inv,jacobian) &
!$OMP          REDUCTION(max:singular_el)
!$OMP DO    
    DO el = 1, myDim_elem2D

       ! LOCAL ELEMENT DEFINITION
       !  T R A N S F O R M A T I O N - M A T R I X   "jacobian"

       jacobian(1:2, 1) = scaling * (coord_nod(1:2, elem_nod(2, el)) - coord_nod(1:2, elem_nod(1, el)))
       jacobian(1:2, 2) = scaling * (coord_nod(1:2, elem_nod(3, el)) - coord_nod(1:2, elem_nod(1, el)))

       if ( coordinate_type == 1 ) then
          !correct for global grids for jumps across -180/180
          if (abs(jacobian(1, 1)) > .5*pi*r_earth) then
             jacobian(1, 1) = -sign(1._dp,jacobian(1, 1))*(2._wp*pi*r_earth-abs(jacobian(1, 1)))
          end if
          if (abs(jacobian(1, 2)) > .5*pi*r_earth) then
             jacobian(1, 2) = -sign(1._dp,jacobian(1, 2))*(2._wp*pi*r_earth-abs(jacobian(1, 2)))
          end if
          jacobian(1, :) = jacobian(1, :) * one_third*sum(cos_nod(elem_nod(1:3,el)))
       END IF
       
       !  I N V E R S E   O F   jacobian
       DET = jacobian(1,1)*jacobian(2,2) - jacobian(1,2)*jacobian(2,1)
       IF ( DET == 0.0 ) THEN
          singular_el = max(singular_el, el)
       ELSE
          DET_inv = 1._dp/DET
       ENDIF

       voltriangle(el) = abs(DET) * std_Vol
       
       !  DEFINITION OF :
       !    - 2D STANDARD ELEMENTS
       !    - BASISFUNCTIONS ON 2D STANDARD ELEMENTS (stdbafu)
       !     Zeta    : stdbafu(1)=1-x-y   stdbafu(2)=x       stdbafu(3)=y
       !     Velocity: stdbafu(1)=1-2y    stdbafu(2)=2x+2y-1 stdbafu(3)=1-2x   
       
       !    derivative_zeta_stdbf      =  0.       
       !    derivative_zeta_stdbf(:,1) = -1.
       !    derivative_zeta_stdbf(1,2) =  1. 
       !    derivative_zeta_stdbf(2,3) =  1.
       
       
       !    derivative_loczeta = MATMUL(TRANSPOSE(jacobian_inv), derivative_zeta_stdbf)
       
       bafu_xy(1, el) = DET_inv*( -jacobian(2,2) + jacobian(2,1))
       bafu_xy(2, el) = DET_inv*(  jacobian(2,2))
       bafu_xy(3, el) = DET_inv*( -jacobian(2,1))

       bafu_xy(4, el) = DET_inv*(  jacobian(1,2) - jacobian(1,1))
       bafu_xy(5, el) = DET_inv*( -jacobian(1,2))
       bafu_xy(6, el) = DET_inv*(  jacobian(1,1))

    ENDDO
!$OMP END DO
!$OMP END PARALLEL

    call com_halo_begin(com_elem2D_full, bafu_xy)
    call com_halo_begin(com_elem2D_full, voltriangle)
    call com_halo_end(com_elem2D_full)
    
    if (singular_el > 0) then
       print *,'Degenerated element (zero volume) in', singular_el
       stop 1
    end if

END SUBROUTINE basisfunctions


