! Part of TsunAWI
! Copyright (c) 2014
! Alfred Wegener Institut,
! Helmholtz Centre for Polar and Marine Research
! Contact: rakowsky@awi.de


!---------------------------------------------------------------------
SUBROUTINE ascii_out_init

    USE PARAMETERS
    USE MESH
    USE ELEMENTS
    USE SWE
  
!---- local declarations

    IMPLICIT NONE
    CHARACTER(len=100)    :: ssh_file, vel_file, ttt_file
    INTEGER               :: i

    if (npes > 1 .and. mype==0) then
       print *,'!!! Sorry, ascii output is not implemented for MPI !!!'
       print *,'!!! with distributed mesh.                         !!!'
       print *,'!!! As MPI is usually a choice for large meshes,   !!!'
       print *,'!!! it is not on our priority list...              !!!'
       return
    end if
    ssh_file   = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'_ssh_000000.out'
    vel_file   = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'_vel_000000.out'

!---- open file & write sea surface height  
  
    OPEN(41, file = ssh_file)
    DO i = 1, nod2D
        WRITE(41,*) real(ssh2(i),sp)
    END DO
    CLOSE(41)

!----- open file & write velocity

    if (write_velocity) then
       OPEN(41, file = vel_file)
       DO i = 1, nod2D
          WRITE(41,*)  real(uv_node(1,i),sp), real(uv_node(2,i),sp)
       END DO
       CLOSE(41)
    endif

    if (write_ttt) then
       ttt_file   = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'_ttt.out'
       OPEN(41, file = ttt_file)
       DO i = 1, nod2D
          WRITE(41, *) real(ttt(i),sp)
       END DO
       CLOSE(41)
    endif

  END SUBROUTINE ascii_out_init
!=====================================
SUBROUTINE ascii_out

    USE PARAMETERS
    USE MESH
    USE ELEMENTS
    USE SWE
  
!---- local declarations

    IMPLICIT NONE
    CHARACTER(len=100)    :: ssh_file, vel_file
    INTEGER               :: i
    INTEGER, save         :: n_out=0
    CHARACTER(len=6)      :: c_out

    if (npes > 1 .and. mype==0) then
       print *,'!!! Sorry, ascii output not implemented for MPI !!!'
       print *,'!!! with distributed mesh.                      !!!'
       return
    end if

    n_out = n_out+1
    write(c_out,'(i6.6)') n_out

    ssh_file   = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'_ssh_'//c_out//'.out'
    vel_file   = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'_vel_'//c_out//'.out'

!---- open file & write sea surface height  
  
    OPEN(41, file = ssh_file)
    DO i = 1, nod2D
        WRITE(41, *) real(ssh2(i),sp)
    END DO
    CLOSE(41)

!----- open file & write velocity

    if (write_velocity) then
       OPEN(41, file = vel_file)
       DO i = 1, nod2D
          WRITE(41,*)  real(uv_node(1,i),sp), real(uv_node(2,i),sp)
       END DO
       CLOSE(41)
    endif

  END SUBROUTINE ascii_out

!================================================================

SUBROUTINE ascii_out_finalize    

  USE PARAMETERS
  USE MESH
  USE ELEMENTS
  USE SWE
  
!---- local declarations

    IMPLICIT NONE
    CHARACTER(len=100)    :: mwh_file, arr_file
    INTEGER               :: i

    if (npes > 1 .and. mype==0) then
       print *,'!!! Sorry, ascii output not implemented for MPI !!!'
       print *,'!!! with distributed mesh.                      !!!'
       return
    end if

    arr_file   = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'_arrivaltime.out'
    mwh_file   = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'_mwh.out'

!---- open file & write arrival time

    OPEN(41, file = arr_file)
    DO i = 1, nod2D
        WRITE(41, '(2e11.4)') arrival_time(i)
    END DO
    CLOSE(41)

!---- open file & write maximum wave height

    OPEN(41, file = mwh_file)
    DO i = 1, nod2D
        WRITE(41, '(e10.3)') mwh(i)
    END DO
    CLOSE(41)


  END SUBROUTINE ascii_out_finalize


!=======================================================================================

SUBROUTINE bin_restart_in

  USE PARAMETERS
  USE MESH
  USE ELEMENTS
  USE SWE
  
  IMPLICIT NONE
  
  INTEGER :: i
  CHARACTER(LEN=100) :: fname
  INTEGER :: restartfile_ssh
  INTEGER :: restartfile_uv
  INTEGER :: restartfile_grad
  INTEGER :: restartfile_diag
  INTEGER :: restartfile_info
  INTEGER :: restartfile_nodhn

    if (npes > 1 .and. mype==0) then
       print *,'!!! Sorry, restart not (yet) implemented for MPI !!!'
       print *,'!!! with distributed mesh.                       !!!'
       return
    end if

  if (enable_nc_out) netcdf_file  = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'.nc'

       fname=TRIM(restart_file)//'_nodhn'
       OPEN(newunit=restartfile_nodhn,file=fname,form='unformatted')
       read(restartfile_nodhn) nodhn(:)
       CLOSE(restartfile_nodhn)

       fname=TRIM(restart_file)//'_ssh'

       OPEN(newunit=restartfile_ssh,file=fname,form='unformatted')
       READ(restartfile_ssh) ssh0(:)
       READ(restartfile_ssh) ssh1(:)
       READ(restartfile_ssh) ssh2(:)
       CLOSE(restartfile_ssh)

       fname=TRIM(restart_file)//'_uv'

       OPEN(newunit=restartfile_uv,file=fname,form='unformatted')
       READ(restartfile_uv) uv0(:,:)
       READ(restartfile_uv) uv1(:,:)
       READ(restartfile_uv) uv2(:,:)
       READ(restartfile_uv) uv_node(:,:)
       CLOSE(restartfile_uv)

       fname=TRIM(restart_file)//'_grad'

       OPEN(newunit=restartfile_grad,file=fname,form='unformatted')
       READ(restartfile_grad) grad(:,:)
       CLOSE(restartfile_grad)


       fname=TRIM(restart_file)//'_diag'

       OPEN(newunit=restartfile_diag, file=fname,form='unformatted')
       READ(restartfile_diag) arrival_time(:)
       READ(restartfile_diag) mwh(:)
       READ(restartfile_diag) max_flux(:)
       READ(restartfile_diag) max_abs_vel(:)
       CLOSE(restartfile_diag)
       
       fname=TRIM(restart_file)//'_info'

       OPEN(newunit=restartfile_info,file=fname,form='unformatted')
       READ(restartfile_info) time0
       CLOSE(restartfile_info)

        time=time0

END SUBROUTINE bin_restart_in

!========================================

SUBROUTINE bin_restart_out

  USE PARAMETERS
  USE MESH
  USE ELEMENTS
  USE SWE
  
  IMPLICIT NONE
  
  INTEGER :: i
  CHARACTER(LEN=100) :: fname
  INTEGER :: restartfile_ssh
  INTEGER :: restartfile_uv
  INTEGER :: restartfile_grad
  INTEGER :: restartfile_diag
  INTEGER :: restartfile_info
  INTEGER :: restartfile_nodhn
  INTEGER :: outfile_status

  if (npes > 1 .and. mype==0) then
     print *,'!!! Sorry, restart not (yet) implemented for MPI !!!'
     print *,'!!! with distributed mesh.                       !!!'
     return
  end if

  fname=TRIM(restart_file)//'_ssh'
  
  OPEN(newunit=restartfile_ssh,file=fname,form='unformatted')
  WRITE(restartfile_ssh) ssh0(:)
  WRITE(restartfile_ssh) ssh1(:)
  WRITE(restartfile_ssh) ssh2(:)
  CLOSE(restartfile_ssh)

  fname=TRIM(restart_file)//'_uv'

  OPEN(newunit=restartfile_uv,file=fname,form='unformatted')
  WRITE(restartfile_uv) uv0(:,:)
  WRITE(restartfile_uv) uv1(:,:)
  WRITE(restartfile_uv) uv2(:,:)
  WRITE(restartfile_uv) uv_node(:,:)
  CLOSE(restartfile_uv)

  fname=TRIM(restart_file)//'_grad'

  OPEN(newunit=restartfile_grad,file=fname,form='unformatted')
  WRITE(restartfile_grad) grad(:,:)
  CLOSE(restartfile_grad)

  fname=TRIM(restart_file)//'_diag'

  OPEN(newunit=restartfile_diag,file=fname,form='unformatted')
  WRITE(restartfile_diag) arrival_time(:)
  WRITE(restartfile_diag) mwh(:)
  WRITE(restartfile_diag) max_flux(:)
  WRITE(restartfile_diag) max_abs_vel(:)
  CLOSE(restartfile_diag)
  
  fname=TRIM(restart_file)//'_info'

  OPEN(newunit=restartfile_info,file=fname,form='unformatted')
  WRITE(restartfile_info) time
  CLOSE(restartfile_info)

  fname=TRIM(restart_file)//'_nodhn'

  OPEN(newunit=restartfile_nodhn,file=fname,form='unformatted')
  WRITE(restartfile_nodhn) nodhn(:)
  CLOSE(restartfile_nodhn)
  
  OPEN(newunit=outfile_status,file='status.txt')
  write(outfile_status,*) 'go on'
  CLOSE(outfile_status)
  
END SUBROUTINE bin_restart_out


!==================================================================================


SUBROUTINE ascii_out_tidegauge_init

  USE PARAMETERS
  USE MESH
  USE SWE

  IMPLICIT NONE

! The mareogram for each TG is extracted from the nearest grid node
! with a water depth of mindepth. If such a node is more than distmax
! away from the TG location, the TG is skipped.   
  real(kind=dp), parameter        :: distmax=10000.  ! in meters (here: 10km)
  real(kind=dp), parameter        :: mindepth=1.  ! in meters (here: 1m)

  real(kind=dp)                   :: mindepth_gg
  integer                        :: n_all
  integer                        :: tg, n, count, io_stat, pos1, pos2
  real(kind=dp), allocatable      :: lon(:), lat(:), dist(:)
  real(kind=dp)                   :: maxlon, minlon, maxlat, minlat
  character(len=20)              :: tg_id
  character(len=1)               :: char
  character(len=100)             ::  dummy 
  character(len=100), allocatable:: tg_urn(:)
  integer, allocatable           :: aux(:)
  real(kind=dp)                   :: r
  character(len=4)                :: land_id


    if (npes > 1 .and. mype==0) then
       print *,'!!! Sorry, tide gauge output not (yet) implemented for MPI !!!'
       print *,'!!! with distributed mesh.                                 !!!'
       return
    end if

  
  ! Define identifier for gauges on land (for inundation studies)
  land_id='LAND';

  open(55,file=TRIM(TideGaugeFile))
  io_stat = 0
  n_all = -1
  DO WHILE (io_stat==0)
     read(55,*,iostat=io_stat) char
     n_all = n_all+1
  END DO
  close(55)
  print *,'Total number of tide gauges: ',n_all

  
  allocate(lon(n_all), lat(n_all))
  allocate(aux(n_all), dist(n_all))
  allocate(tg_urn(n_all))

  open(55,file=TRIM(TideGaugeFile))
  DO tg = 1, n_all
     read(55,*) tg_urn(tg), lon(tg), lat(tg)
  END DO
  close(55)

! Find nearest grid node in water. If the distance between
! node and TG location is larger than distmax, reject tidegauge 

  aux(:)  = 0
  dist(:) = 4.*pi*pi
  count = 0

! computing grid bounding box
  maxlon = (maxval(coord_nod(1,:)) + distmax/r_earth) * rad_inv
  minlon = (minval(coord_nod(1,:)) - distmax/r_earth) * rad_inv
  maxlat = (maxval(coord_nod(2,:)) + distmax/r_earth) * rad_inv
  minlat = (minval(coord_nod(2,:)) - distmax/r_earth) * rad_inv

  DO tg = 1, n_all
     if (lat(tg) > minlat .and. lat(tg) < maxlat .and. &
         lon(tg) > minlon .and. lon(tg) < maxlon) then

        mindepth_gg=mindepth
        if (tg_urn(tg)(1:4)==land_id) then
           print *,'Land gauge'
           mindepth_gg=-100.0
        end if

        DO n=1,nod2D
           if (nodhn(n) >= mindepth_gg) then
              ! Remind to convert lon, lat to radian
              r = (rad*lon(tg) - coord_nod(1,n))**2 + (rad*lat(tg) - coord_nod(2,n))**2 
              if ( r < dist(tg)) then
                 aux(tg) = n
                 dist(tg) = r
              endif
           end if
        ENDDO
     
        dist(tg) = sqrt(dist(tg)) *r_earth
        
        if (dist(tg) <= distmax) then
           count = count+1
           if (verbosity >= 5) &
                print *,trim(tg_urn(tg)),': nearest sea node ',aux(tg),', distance ',dist(tg),'m'
        elseif (verbosity >= 5) then
                print *,'Tidegauge ',trim(tg_urn(tg)),' not considered: distance [m] to nearest sea node',dist(tg)
        endif

     elseif (verbosity >= 5) then
        dist(tg) = sqrt(dist(tg)) *r_earth
        print *,'Tidegauge ',trim(tg_urn(tg)),' not considered: outside bounding box of the computational domain'
     endif
     
  ENDDO

  n_tidegauges = count

 if (verbosity >= 1)  print *,'Number of tide gauges taken into account: ',n_tidegauges
  allocate(tidegauge_node_index(n_tidegauges))
  allocate(tidegauge_ascii_filename(n_tidegauges))

  n = 0
  DO tg = 1,n_all
     if (dist(tg) <= distmax) then
        n=n+1

        dummy = tg_urn(tg)
        if (LEN_TRIM(dummy)>8) then
          pos1 = index(tg_urn(tg),':',back=.true.) +1
          pos2 = index(tg_urn(tg),' ',back=.false.) -1
          tg_id = dummy(pos1:pos2)
        else
          tg_id = trim(dummy)
       end if

        tidegauge_node_index(n) = aux(tg)
        tidegauge_ascii_filename(n) = 'TG_'//trim(tg_id)//'_'//trim(TSID)//'.out'

        if (verbosity >= 5) &
             print *, trim(tidegauge_ascii_filename(n))

        if (.not. read_restart) then
           ! write the information in the preambel
           open(66,file=tidegauge_ascii_filename(n))
           write(66,'(a2$)') "% "
           write(66,*) trim(tg_urn(tg))
           write(66,'("% TG longitude, latitude      ",2f12.6)') lon(tg), lat(tg)
           if (tg_urn(tg)(1:4)==land_id) then
              write(66,'("% Index of nearest land grid node ",i12)') tidegauge_node_index(n)
           else 
              write(66,'("% Index of nearest grid node with depth>1m",i12)') tidegauge_node_index(n)
           end if
           write(66,'("% Grid node longitude latitude",2f12.6)') coord_nod(1:2,tidegauge_node_index(n))/rad
           write(66,'("% Distance [m] to grid node",f12.4)') dist(tg)
           if (tg_urn(tg)(1:4)==land_id) then
              write(66,'("% Land height [m] at grid node before rupture",f12.4)') &
                   -nodhn_init(tidegauge_node_index(n))
              write(66,'("% Land height [m] at grid node after rupture ",f12.4)') &
                   -nodhn(tidegauge_node_index(n))
           else
              write(66,'("% Water depth [m] at grid node before rupture",f12.4)') &
                   nodhn_init(tidegauge_node_index(n))
              write(66,'("% Water depth [m] at grid node after rupture ",f12.4)') &
                   nodhn(tidegauge_node_index(n))
           end if
           write(66,'(a52)') "% time since rupture [min]   sea surface elevation [m]"
           write(66,'(f15.2,f13.5)') time/60., ssh2(tidegauge_node_index(n))        
           close(66)
        endif
     endif
  ENDDO


END SUBROUTINE ascii_out_tidegauge_init


!==================================================================================


SUBROUTINE ascii_out_tidegauge

  USE PARAMETERS
  USE MESH
  USE SWE

  IMPLICIT NONE

  integer :: tg

    if (npes > 1 .and. mype==0) then
       print *,'!!! Sorry, tide gauge output not (yet) implemented for MPI !!!'
       print *,'!!! with distributed mesh.                                 !!!'
       return
    end if

  DO tg = 1, n_tidegauges
     open(66,file=tidegauge_ascii_filename(tg),status='old',position='append',action='write')
     write(66,'(f15.2,f13.5)') time/60., ssh2(tidegauge_node_index(tg))
     close(66)
  ENDDO


END SUBROUTINE ascii_out_tidegauge


!====================================================================================

SUBROUTINE write_checksums

  USE PARAMETERS
  USE MESH
  USE ELEMENTS
  USE SWE
  
!---- local declarations

  IMPLICIT NONE
  CHARACTER(len=100)    :: checksum_file
  integer               :: u_checksum
  
  real(kind=wp)         :: cs_max(3)
  real(kind=wp)         :: cs_min(3)
  real(kind=wp)         :: cs_sum(3)
  integer               :: n, ed

  cs_max(:) = -1000._wp
  cs_min(:) =  1000._wp
  cs_sum(:) = 0._wp

!$OMP PARALLEL DEFAULT(shared) PRIVATE(n,ed) &
!$OMP          REDUCTION(+:  cs_sum) &
!$OMP          REDUCTION(max:cs_max) &
!$OMP          REDUCTION(min:cs_min) 

!$OMP DO 
  do n=1,myDim_nod2D
     cs_max(1) = max(cs_max(1),ssh2(n))
     cs_min(1) = min(cs_min(1),ssh2(n))
     cs_sum(1) =     cs_sum(1)+ssh2(n)
  enddo
!$OMP END DO NOWAIT
!$OMP DO 
  do ed=1,myDim_edg2D
     cs_max(2) = max(cs_max(2),uv2(1,ed));  cs_max(3) = max(cs_max(3),uv2(2,ed)); 
     cs_min(2) = min(cs_min(2),uv2(1,ed));  cs_min(3) = min(cs_min(3),uv2(2,ed)); 
     cs_sum(2) =     cs_sum(2)+uv2(1,ed);   cs_sum(3) =     cs_sum(3)+uv2(2,ed);     
  enddo
!$OMP END DO
!$OMP END PARALLEL

  call com_reduce_begin(cs_max, com_max)
  call com_reduce_begin(cs_min, com_min)
  call com_reduce_begin(cs_sum, com_sum)
  call com_reduce_end

  if (mype==0) then
     checksum_file   = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'_checksums.out'
  
     open(newunit=u_checksum, file=checksum_file, status='replace', action='write')
  
     write(u_checksum,*) "max_ssh2=",cs_max(1)
     write(u_checksum,*) "min_ssh2=",cs_min(1)
     write(u_checksum,*) "sum_ssh2=",cs_sum(1)
     write(u_checksum,*) "max_u2=  ",cs_max(2)
     write(u_checksum,*) "min_u2=  ",cs_min(2)
     write(u_checksum,*) "sum_u2=  ",cs_sum(2)
     write(u_checksum,*) "max_v2=  ",cs_max(3)
     write(u_checksum,*) "min_v2=  ",cs_min(3)
     write(u_checksum,*) "sum_v2=  ",cs_sum(3)
     
     close(u_checksum)
  endif
  
END SUBROUTINE write_checksums
