! Part of TsunAWI
! Copyright (c) 2014
! Alfred Wegener Institut,
! Helmholtz Centre for Polar and Marine Research
! Contact: rakowsky@awi.de



#ifdef NO_NETCDF
! Dummy routines, if no netcdf library is available

  SUBROUTINE nc_init
  END SUBROUTINE nc_init
  
  SUBROUTINE nc_out(i)
    integer, intent(in) :: i
  END SUBROUTINE nc_out

  SUBROUTINE nc_finalize
  END SUBROUTINE nc_finalize

#else
  module netcdf_variables
    use parameters
    implicit none
    include 'netcdf.inc'
    
    INTEGER 		     :: Fid
    INTEGER 		     :: VarId_iter, VarId_time ! numbers
    INTEGER 		     :: VarId_topo             ! topography (after rupture and smoothing)
    INTEGER 		     :: VarId_topo_ini         ! inital topography
    INTEGER 		     :: VarId_ssh              ! fields: ssh
    INTEGER 		     :: VarId_arrival          ! fields: Arrival times
    INTEGER 		     :: VarId_mwh              ! fields: maximum wave height
    INTEGER 		     :: VarId_flux, VarId_vel  !fields: flux and mean absolute velocity
    INTEGER		     :: VarId_nodal_vel 	   !fields: vector
    INTEGER		     :: VarId_nodal_cfl 	   !fields: cfl value
    INTEGER		     :: VarId_epot, VarId_ekin !fields: energies
    INTEGER                  :: VarId_trian		     ! variable: triangle
    INTEGER                  :: VarId_lon, VarId_lat          ! variables: longitude, latitude
    INTEGER                  :: VarId_tri_area		     ! variable: area
    INTEGER                  :: VarId_uplift                  ! variable: topo/bathy uplift
    INTEGER                  :: VarId_index	             ! variables: topography, index
    REAL(kind=dp), parameter :: notyetknown=-99999.

  end module netcdf_variables

  
SUBROUTINE nc_init

!----- uses

    USE MESH
    USE ELEMENTS
    USE SWE
    USE OKADA_FAULT_PARAMETERS
    use netcdf_variables
!-----local declarations
    implicit none


    CHARACTER(len=100)      :: ncfile
    CHARACTER(len=200)      :: attstr
    CHARACTER(len=2)        :: par_num
    CHARACTER(len=3)        :: StrMeshID
    INTEGER                 :: i
    INTEGER                 :: DimId_iter		     ! dimension: iteration
    INTEGER                 :: DimId_n2D		     ! dimension: nodes
    INTEGER                 :: DimId_el2D		     ! dimension: elements
    INTEGER                 :: Dim1, Dim2, Dim3, Dim4	     ! dimensions: 
    INTEGER                 :: VarId_ttt                     ! variable: estimated arrival time ttt
    INTEGER                 :: s       			     ! auxiliary: status counter
    INTEGER, DIMENSION(3)   :: dimarray			     ! auxiliary: array dimension
    INTEGER, DIMENSION(500) :: stat    			     ! auxiliary: status array
        
    real(kind=sp), allocatable :: global1(:)
    REAL(KIND=dp)              ::   X0, Y0, D, L, W, TH, DL, HH, RD

    REAL(KIND=dp)              :: att_mw
    integer                    :: att_epi_i,  att_epi_j, n, el

    if (write_diag_in_nc_snapshots .and. npes>1) then
       write_diag_in_nc_snapshots = .false.
       if (mype==0) &
            print *,'/model_output/  write_diag_in_nc_snapshots ignored for MPI parallel code'
     endif
     
    if (mype==0) then
    ! Set nc output filename

    netcdf_file  = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'.nc'

!----- open file and write global attributes

    s = 1;  stat(s) = NF_CREATE(TRIM(netcdf_file), 0, fid) 
    
    attstr  = 'Tsunami Simulation'
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'title',       LEN_TRIM(attstr),    &
                              TRIM(attstr)) 

    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'institution', 3, 'AWI' ) 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'Conventions', 6, 'CF-1.0') 

    IF (enable_ruptgen_scenario) THEN

       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'TSID',     LEN_TRIM(TSID), TSID) 
       
       attstr  = 'RuptGen V2.0 - Rupture Generator provided by GFZ Potsdam, Germany'
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'sourceDescription', LEN_TRIM(attstr),    &
            TRIM(attstr)) 
       s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Input_Magnitude',         NF_DOUBLE, 1, real(eq_mag,dp)) 
       s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Input_Epicenter_X',       NF_DOUBLE, 1, real(eq_epi_x,dp)) 
       s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Input_Epicenter_Y',       NF_DOUBLE, 1, real(eq_epi_y,dp)) 
       ! s=s+1;
       !stat(s) = NF_PUT_ATT_INT(fid,    NF_GLOBAL, 'Tsunami_Input_Direction_Rupture', NF_FLOAT,  1, eq_dir_rupt) 
       s=s+1; stat(s) = NF_PUT_ATT_INT(fid,    NF_GLOBAL, 'sourceParameterRuptGen_i',        NF_INT,    1, Epi_i) 
       s=s+1; stat(s) = NF_PUT_ATT_INT(fid,    NF_GLOBAL, 'sourceParameterRuptGen_j',        NF_INT,    1, Epi_j)    
       s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'sourceParameterRuptGen_mw',       NF_DOUBLE, 1, real(mw,dp)) 
       attstr  = 'Topography data: SRTM X- and C-Band data Interpolated to X-Band resolution, adjusted in Kuta/Bali by local data'
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid,   NF_GLOBAL, 'meshTopographyDataDescription', LEN_TRIM(attstr),    &
             TRIM(attstr))
       attstr  = 'Bathymetry data: GEBCO 30sec, SONNE Cruise SO 186, SCOTT Cruise, Seamap Data; Padang Data by Franzius Institut' 
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid,   NF_GLOBAL, 'meshBathymetryDataDescription', LEN_TRIM(attstr),    &
             TRIM(attstr))

    ELSE IF (enable_okada_scenario) THEN
 
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'sourceDescription', 16, 'Okada parameters')
       s=s+1; stat(s) = NF_PUT_ATT_INT(fid,  NF_GLOBAL,  'number_of_fault_planes',          NF_INT,    1, nfault) 
             
       IF (nfault == 1) THEN
           X0 = okada_faults(1)%lon   
           Y0 = okada_faults(1)%lat   
           D  = okada_faults(1)%slip  
           L  = okada_faults(1)%len   
           W  = okada_faults(1)%width 
           TH = okada_faults(1)%strike
           DL = okada_faults(1)%dip   
           HH = okada_faults(1)%depth 
           RD = okada_faults(1)%rake  
                    

          s=s+1; stat(s)= NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Plane_X0',             NF_DOUBLE, 1, X0) 
          s=s+1; stat(s)= NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Plane_Y0',             NF_DOUBLE, 1, Y0) 
          s=s+1; stat(s)= NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Slip',                 NF_DOUBLE, 1,  D) 
          s=s+1; stat(s)= NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Fault_Length',         NF_DOUBLE, 1,  L) 
          s=s+1; stat(s)= NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Fault_Width',          NF_DOUBLE, 1,  W) 
          s=s+1; stat(s)= NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Strike_Angle',         NF_DOUBLE, 1, TH) 
          s=s+1; stat(s)= NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Dip_Angle',            NF_DOUBLE, 1, DL) 
          s=s+1; stat(s)= NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Fault_Depth',          NF_DOUBLE, 1, HH) 
          s=s+1; stat(s)= NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Direction_Dislocation',NF_DOUBLE, 1, RD) 

       ELSEIF (nfault <= 20) then
          DO i = 1, nfault
             WRITE(par_num, '(I2.2)') i
             X0 = okada_faults(i)%lon   
             Y0 = okada_faults(i)%lat   
             D  = okada_faults(i)%slip  
             L  = okada_faults(i)%len   
             W  = okada_faults(i)%width 
             TH = okada_faults(i)%strike
             DL = okada_faults(i)%dip   
             HH = okada_faults(i)%depth 
             RD = okada_faults(i)%rake
             
             s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Plane_X0_'//par_num,              &
                   NF_DOUBLE, 1, X0) 
             s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Plane_Y0_'//par_num,              &
                   NF_DOUBLE, 1, Y0) 
             s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Slip_'//par_num,                  &
                   NF_DOUBLE, 1, D) 
             s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Fault_Length_'//par_num,          &
                   NF_DOUBLE, 1, L) 
             s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Fault_Width_'//par_num,           &
                   NF_DOUBLE, 1, W) 
             s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Strike_Angle_'//par_num,          &
                   NF_DOUBLE, 1, TH) 
             s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Dip_Angle_'//par_num,             &
                   NF_DOUBLE, 1, DL)
             s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Fault_Depth_'//par_num,           &
                   NF_DOUBLE, 1, HH) 
             s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'Tsunami_Okada_Direction_Dislocation_'//par_num, &
                   NF_DOUBLE, 1, RD) 
          END DO
      END IF

   elseif (enable_aifdr_scenario) then

       attstr  = 'Source provided by AIFDR'
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid,   NF_GLOBAL, 'sourceDescription', len_trim(attstr), &
                                 TRIM(attstr))
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid,   NF_GLOBAL, 'source_zone',  len_trim(source_zone),&
            TRIM(source_zone))
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid,   NF_GLOBAL, 'source_zone_id', len_trim(sz),     sz)
       s=s+1; stat(s) = NF_PUT_ATT_INT(fid,    NF_GLOBAL, 'sourceParameterEpi_i',  NF_INT,    1, epi_i)    
       s=s+1; stat(s) = NF_PUT_ATT_INT(fid,    NF_GLOBAL, 'sourceParameter_Epi_j', NF_INT,    1, epi_j)    
       s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'sourceParameter_mw',    NF_DOUBLE, 1, real(mw,dp))

    elseif (enable_idealised) then

       attstr  = 'Source: Idealised, cosine bell'
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid,   NF_GLOBAL, 'sourceDescription', len_trim(attstr),    &
                                 TRIM(attstr))

       s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL,  'sourceParameterEpi_lon',  NF_DOUBLE, 1, real(eq_epi_x,dp))    
       s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL,  'sourceParameter_Epi_lat', NF_DOUBLE, 1, real(eq_epi_y,dp))    
       s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'sourceParameter_mag',      NF_DOUBLE, 1, real(eq_mag,dp))

    END IF

    WRITE(StrMeshID, '(I3.3)') MeshID
    attstr  = 'TsunAWI Revision '//revision//' - Mesh: '//StrMeshID
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'source', LEN_TRIM(attstr),     &
                              TRIM(attstr)) 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'history',      8, 'no entry') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'references',   4, 'none') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'comment',      4, 'none') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'provider',     3, 'AWI')
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'model',        7, 'TsunAWI')
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'modelVersion', LEN_TRIM(revision), TRIM(revision)) 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'scope',        8, 'regional')
    
    s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'globalMinSSH',              NF_DOUBLE,1, notyetknown) 
    s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'globalMaxSSH',              NF_DOUBLE,1, notyetknown) 
    s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'boundingBox_UpperRight_Lon',NF_DOUBLE,1, real(BoundingBox_xmax,dp)) 
    s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'boundingBox_UpperRight_Lat',NF_DOUBLE,1, real(BoundingBox_ymax,dp)) 
    s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'boundingBox_LowerLeft_Lon', NF_DOUBLE,1, real(BoundingBox_xmin,dp)) 
    s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'boundingBox_LowerLeft_Lat', NF_DOUBLE,1, real(BoundingBox_ymin,dp)) 
    s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'outputTimestepSize',        NF_DOUBLE,1, real(T_out,dp)) 
    s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'integrationTime',           NF_DOUBLE,1, real(T_end,dp))
    s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'spatialResolution_Coarse',  NF_DOUBLE,1, real(longestEdge,dp))
    s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'spatialResolution_Fine',    NF_DOUBLE,1, real(shortestEdge,dp))
    s=s+1; stat(s) = NF_PUT_ATT_INT(fid,    NF_GLOBAL,  'degreeOfFreedom',          NF_INT,   1, nod2D)

    if (enable_benchmark .and. benchmark_ident<=2) then
       
       if  (benchmark_ident==1) attstr  = 'Bathymetry / topography data: Sloping beach experiment'
       if  (benchmark_ident==2) attstr  = 'Bathymetry / topography data: Monai beach channel experiment'

       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, NF_GLOBAL, 'meshDataDescription', LEN_TRIM(attstr),    &
            TRIM(attstr)) 
    endif
    



!----- DEFINE DIMENSIONS ------------------------------

    s=s+1; stat(s) = NF_DEF_DIM(fid, 'nodes_2D',    nod2D,        DimId_n2D)             
    s=s+1; stat(s) = NF_DEF_DIM(fid, 'elements_2D', elem2D,       DimId_el2D)        
    s=s+1; stat(s) = NF_DEF_DIM(fid, 'one',         1,            dim1)                           
    s=s+1; stat(s) = NF_DEF_DIM(fid, 'two',         2,            dim2)                           
    s=s+1; stat(s) = NF_DEF_DIM(fid, 'three',       3,            dim3)                         
    s=s+1; stat(s) = NF_DEF_DIM(fid, 'four',        4,            dim4)                          
    s=s+1; stat(s) = NF_DEF_DIM(fid, 'iteration',   NF_UNLIMITED, DimId_iter)

    DO i = 1,  s
       IF (stat(i) /= NF_NOERR) THEN
          PRINT *, 'ncinit, NetCDF error',stat(i),' in dimension definitions, no.', i
          print *, nf_strerror(stat(i))
       endif
    END DO
    s = 0

!----- DEFINE VARIABLES ---------------------------------
!----- coordinates
    s=s+1; stat(s) = NF_DEF_VAR(fid, 'longitude', NF_FLOAT, 1, DimId_n2D, VarId_lon) 
    s=s+1; stat(s) = NF_DEF_VAR(fid, 'latitude',  NF_FLOAT, 1, DimId_n2D, VarId_lat) 
    s=s+1;

!----- grid

    dimarray = (/ dim3, DimId_el2D, 0 /)

    s=s+1; stat(s) = NF_DEF_VAR(fid, 'triangles',  NF_INT,   2, dimarray(1:2), VarId_trian) 
    s=s+1; stat(s) = NF_DEF_VAR(fid, 'tri_area',   NF_FLOAT, 1, DimId_el2D,    VarId_tri_area) 
    
!----- node indices
    s=s+1; stat(s) = NF_DEF_VAR(fid, 'node_index', NF_INT,   1, DimId_n2D,     VarId_index) 
    
!----- numbers 
    s=s+1; stat(s) = NF_DEF_VAR(fid, 'iteration',  NF_INT,   1, DimId_iter,    VarId_iter) 
    s=s+1; stat(s) = NF_DEF_VAR(fid, 'time',       NF_FLOAT, 1, DimId_iter,    VarId_time) 

!----- energy
    if (write_energy) then
       s=s+1; stat(s) = NF_DEF_VAR(fid, 'epot',    NF_FLOAT, 1, DimId_iter,    VarId_epot) 
       s=s+1; stat(s) = NF_DEF_VAR(fid, 'ekin',    NF_FLOAT, 1, DimId_iter,    VarId_ekin) 
    endif
       
!----- topography
    s=s+1; stat(s) = NF_DEF_VAR(fid,'initial_topography',NF_FLOAT, 1, DimId_n2D, VarId_topo_ini) 
    s=s+1; stat(s) = NF_DEF_VAR(fid,'topography',  NF_FLOAT, 1, DimId_n2D,     VarId_topo) 


!----- F I E L D S 

!-----  scalar variables

    dimarray = (/ DimId_n2D, DimId_iter, 0 /)
    s=s+1; stat(s) = NF_DEF_VAR(fid, 'ssh', NF_FLOAT, 2, dimarray(1:2), VarId_ssh); 

!----- arrival times, max. wave height, max. abs. velocity, max. flux, initial uplift

    s=s+1; stat(s) = NF_DEF_VAR(fid, 'first_arrival',     NF_FLOAT, 1, DimId_n2D, VarId_arrival)
 
    if (write_ttt) &
       s=s+1; stat(s) = NF_DEF_VAR(fid, 'first_arrival_ttt', NF_FLOAT, 1, DimId_n2D, VarId_ttt) 

    s=s+1; stat(s) = NF_DEF_VAR(fid, 'max_wave_height',   NF_FLOAT, 1, DimId_n2D, VarId_mwh) 
    s=s+1; stat(s) = NF_DEF_VAR(fid, 'max_abs_vel',       NF_FLOAT, 1, DimId_n2D, VarId_vel) 
    s=s+1; stat(s) = NF_DEF_VAR(fid, 'max_flux',          NF_FLOAT, 1, DimId_n2D, VarId_flux) 
    s=s+1; stat(s) = NF_DEF_VAR(fid, 'initial_uplift',    NF_FLOAT, 1, DimId_n2D, VarId_uplift) 

!----- vector variables

!-----  velocity in nodes
    IF (write_velocity) THEN
       dimarray = (/ dim2, DimId_n2D, DimId_iter /)
       s=s+1; stat(s) = NF_DEF_VAR(fid, 'velocity_in_nodes', NF_FLOAT, 3, dimarray(1:3), VarId_nodal_vel)
    END IF

    IF (write_cfl) THEN
       dimarray = (/ DimId_n2D, DimId_iter, 0 /)
       s=s+1; stat(s) = NF_DEF_VAR(fid, 'cfl', NF_FLOAT, 2, dimarray(1:2), VarId_nodal_cfl)
       
       attstr = 'cfl in nodes, velocity vs egdlen/dt'
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_nodal_cfl, 'long_name', LEN_TRIM(attstr), TRIM(attstr))
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_nodal_cfl, 'field',       19, 'cfl, scalar, series') 
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_nodal_cfl, 'connections', 20, 'triangles, triangles') 
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_nodal_cfl, 'positions',   17, 'surface_locations')

       ! no units, because the CFL is dimensionless
    END IF
 
    DO i = 1,  s
       IF (stat(i) /= NF_NOERR) then
          PRINT *,  'nc_init, NetCDF error',stat(i),' in variable definition, no.', i
          print *, nf_strerror(stat(i))
       end IF
    END DO

!----- DEFINE ATTRIBUTES -----------------------------------------

    s = 0

    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_lon,      'long_name',    9, 'longitude') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_lon,      'units',       12, 'degrees_east') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_lat,      'long_name',    8, 'latitude') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_lat,      'units',       13, 'degrees_north') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_index,    'long_name',   16, 'nodal attributes') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_tri_area, 'long_name',   13, 'triangle area') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ssh,      'long_name',   21, 'sea surface elevation') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ssh,      'units',        5, 'meter') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ssh,      'field',       19, 'ssh, scalar, series') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ssh,      'connections', 20, 'triangles, triangles') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ssh,      'positions',   17, 'surface_locations') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_topo_ini, 'long_name',   18, 'initial_topography') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_topo_ini, 'field',       18, 'initial_topography') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_topo_ini, 'connections', 20, 'triangles, triangles') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_topo_ini, 'positions',   17, 'surface_locations') 
    attstr  = 'meters (before shock bathymetry and topography)'
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_topo_ini, 'units',       LEN_TRIM(attstr), attstr) 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_topo,     'long_name',   10, 'topography') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_topo,     'field',       10, 'topography') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_topo,     'connections', 20, 'triangles, triangles') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_topo,     'positions',   17, 'surface_locations')     
    attstr  = 'meters (after shock bathymetry and topography)'
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_topo,     'units',       LEN_TRIM(attstr), attstr) 
    attstr='initial uplift due to earthquake'
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_uplift,   'long_name',   LEN_TRIM(attstr), attstr) 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_uplift,   'field',       14, 'initial_uplift') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_uplift,   'connections', 20, 'triangles, triangles') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_uplift,   'positions',   17, 'surface_locations') 
    attstr  = 'first wave arrival (sea surface elevation larger than +0.01m)'
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_arrival,  'long_name',   LEN_TRIM(attstr), attstr) 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_arrival,  'eta_type',    12,'eta_thd_orig')
    attstr  = 'seconds after rupture'
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_arrival,  'units',       LEN_TRIM(attstr), attstr) 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_arrival,  'field',       13, 'first_arrival') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_arrival,  'connections', 20, 'triangles, triangles') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_arrival,  'positions',   17, 'surface_locations') 

    if (write_ttt) then
       attstr  = 'first wave arrival (estimated arrival time - ttt)'
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ttt,   'long_name',   LEN_TRIM(attstr), attstr) 
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ttt,   'eta_type',    12, 'ttt_est_orig')
       attstr  = 'seconds after rupture'
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ttt,   'units',       LEN_TRIM(attstr), attstr) 
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ttt,   'field',       17, 'first_arrival_ttt') 
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ttt,   'connections', 20, 'triangles, triangles') 
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ttt,   'positions',   17, 'surface_locations') 
    endif

    attstr  = 'maximum wave height\n (above mean sea level on sea \n and above after shock topography on land)'
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_mwh,       'long_name',   LEN_TRIM(attstr), attstr) 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_mwh,       'units',        6, 'meters') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_mwh,       'field',       15, 'max_wave_height') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_mwh,       'connections', 20, 'triangles, triangles') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_mwh,       'positions',   17, 'surface_locations') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_vel,       'long_name',   36, 'maximum nodal velocity approximation') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_vel,       'units',        5, 'm/sec') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_vel,       'field',       11, 'max_abs_vel')
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_vel,       'connections', 20, 'triangles, triangles') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_vel,       'positions',   17, 'surface_locations') 
    attstr  = 'maximum flux in nodes\n (maximum value of the product of velocity and wave height)'
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_flux,      'long_name',   LEN_TRIM(attstr), attstr) 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_flux,     'units',        7, 'm^2/sec') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_flux,     'field',         8, 'max_flux') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_flux,     'connections',  20, 'triangles, triangles') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_flux,     'positions',    17, 'surface_locations') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_time,     'long_name',     4, 'time'); 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_time,     'units',        19, 'seconds since rupture') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_time,     'time_origin',   1, '0') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_time,     'field',        12, 'time, series') 
    s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_iter,     'long_name',     9, 'iteration') 


    if (write_energy) then
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_epot,   'units',        6, 'Joules') 
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ekin,   'units',        6, 'Joules') 
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_epot,   'long_name',   22, 'total potential energy')
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ekin,   'long_name',   20, 'total kinetic energy')
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_epot,   'field',       14, 'energy, series')
       s=s+1; stat(s) = NF_PUT_ATT_TEXT(fid, VarId_ekin,   'field',       14, 'energy, series')
    endif

    stat(s) = NF_ENDDEF(fid) 
    
    DO i = 1, s
       IF (stat(i) /= NF_NOERR) then
          PRINT *,  'ncinit, NetCDF error',stat(i),' in attribute assignments, no.', i
          print *, nf_strerror(stat(i))
       end IF
    END DO

 endif ! mype==0

 
 if (npes == 1 .and. write_ttt) then
    stat(1) = NF_PUT_VAR_REAL(fid, VarId_ttt,   ttt(:))
    IF (stat(1) /= NF_NOERR) then
       PRINT *,  'ncinit, NetCDF error',stat(1),' writing ttt values.'
       print *, nf_strerror(stat(1))
    end IF
 end if
 
END SUBROUTINE nc_init

SUBROUTINE nc_out(iteration)

!----- uses

    USE MESH
    USE ELEMENTS
    USE SWE
    use netcdf_variables
    use mpi_communication
!----- local declarations
    
    implicit none
    
    INTEGER, INTENT(in) 	 :: iteration
    INTEGER, SAVE 	         :: iter=1

    INTEGER 		         :: i, n, iA, iZ, edg, j
    INTEGER 		         :: pos1, nmb
    INTEGER, DIMENSION(2)      :: pos1vec, nmbvec
    INTEGER, DIMENSION(3)      :: pos1vec3, nmbvec3
 
    INTEGER,    DIMENSION(50)    :: stat                     ! auxiliaries
    INTEGER                      :: s                        ! auxiliaries
    INTEGER                      :: restart_iter, dimid_iter
    REAL(kind=sp)                :: AUX_n2D(myDim_nod2D)
    REAL(kind=sp), allocatable   :: global(:), global2(:,:)
    REAL(kind=sp)                :: acc, inv_acc

    if (mype==0) then
       IF (read_restart .AND. iter == 1) THEN
          stat(1) = NF_OPEN(TRIM(netcdf_file), NF_WRITE, Fid)
          
          IF (stat(1) /= NF_NOERR) then
             print *,'nc_out, nc-file error:',stat(1)
             print *, nf_strerror(stat(1))
             STOP 1
          endif
          stat(2) = NF_INQ_DIMID(fid,  'iteration', dimid_iter)
          stat(2) = NF_INQ_DIMLEN(fid, dimid_iter,  restart_iter)
          iter    = restart_iter + 1
       END IF

!----- WRITE VARIABLES
       s = 0
       s=s+1; stat(s) = NF_PUT_VARA_INT( fid, VarId_iter, iter, 1, iteration) 
       s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_time, iter, 1, REAL(time, sp)) 

       if (write_energy) then
          s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_epot, iter, 1, REAL(epot, sp)) 
          s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_ekin, iter, 1, REAL(ekin, sp)) 
       endif
       
!----- 2D FIELDS
       pos1vec = (/ 1, iter  /)
       nmbvec  = (/ nod2D, 1 /)
    endif
!----- sea surface height
!----- subtract terrain height on land
  
    if (nc_snapshot_accuracy_ssh <= 0.) then
!$OMP PARALLEL DO 
       DO n = 1, myDim_nod2D
          IF (nodhn(n) < 0.) THEN
             AUX_n2D(n) = max(0., ssh1(n) + nodhn(n))
          ELSE
             AUX_n2D(n) = ssh1(n)
          END IF
       END DO
!$OMP END PARALLEL DO
    else
       acc = real(nc_snapshot_accuracy_ssh ,sp)
       inv_acc = 1._sp/nc_snapshot_accuracy_ssh 
!$OMP PARALLEL DO 
       DO n = 1, myDim_nod2D
          IF (nodhn(n) < 0.) THEN
             AUX_n2D(n) =  acc* INT(inv_acc*max(0., real(ssh1(n) + nodhn(n),sp)))
          ELSE
             AUX_n2D(n) =  acc* INT(inv_acc*real(ssh1(n),sp))
          END IF
       END DO
!$OMP END PARALLEL DO
    end if
       
    if (npes == 1) then
       stat(s) = NF_PUT_VARA_REAL(fid, VarId_ssh, pos1vec, nmbvec, AUX_n2D(:)) 

       if (write_diag_in_nc_snapshots) then
          s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_arrival, 1, nod2D, arrival_time(:)) 
          s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_mwh,     1, nod2D, mwh(:))
          s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_vel,     1, nod2D, max_abs_vel(:))
          s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_flux,    1, nod2D, max_flux(:))
       endif

    else ! npes > 1
       if (mype==0) then
          allocate(global(nod2D))
       else
          allocate(global(1))
       endif
       call com_gather(com_nod2D, AUX_n2D, global)
       if (mype==0) then
          s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_ssh, pos1vec, nmbvec, global)
       endif
    end if
    
    IF (write_cfl) THEN
!$OMP PARALLEL DO DEFAULT(SHARED) &
!$OMP&            PRIVATE(n,iA,iZ,j,edg)     
       do n = 1,myDim_nod2D
          AUX_n2D(n)=0.
          if (depth(n) > small) then

             iA = nodedge_ptr(n)
             iZ = nodedge_ptr(n+1)-1
             ! The CFL is calculated on the edges and interpolated to the nodes:
             ! |vel|*dt / |edge|
             ! edg_length contains 1./edg_length here.
             do j=iA, iZ
                edg = nodedge_addr(j)
                AUX_n2D(n) = AUX_n2D(n) + nodedge_wght(j)* &
                     sqrt(uv2(1,edg)*uv2(1,edg) + uv2(2,edg)*uv2(2,edg)) &
                     * edg_length(edg)                      
             end do
             AUX_n2D(n) = AUX_n2D(n) *dt
             
          end if
       end do
!$OMP END PARALLEL DO
       if (npes == 1) then
          s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_nodal_cfl, pos1vec, nmbvec, AUX_n2D)
       else
          call com_gather(com_nod2D, AUX_n2D, global)
          if (mype==0) then
             s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_nodal_cfl, pos1vec, nmbvec, global)
          END IF
       end if
    end IF

    if (npes > 1) deallocate(global)
    
!----- nodal velocity
    IF (write_velocity) THEN       
       pos1vec3 = (/ 1, 1, iter /)
       nmbvec3  = (/ 2, nod2D, 1 /)
       if (npes==1) then
          s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_nodal_vel, pos1vec3, nmbvec3, uv_node(1,1:nod2D))
       else
          if (mype==0) then
             allocate(global2(2,nod2D))
          else
             allocate(global2(1,1))
          endif
          call com_gather(com_nod2D,uv_node,global2)
          if (mype==0) then
             s=s+1; stat(s) = NF_PUT_VARA_REAL(fid, VarId_nodal_vel, pos1vec3, nmbvec3, global2)
          end if
          deallocate(global2)
       end IF
    END IF

    if (mype==0) then
       DO i = 1, s 
          IF (stat(i) /= NF_NOERR) then
             PRINT *, 'nc_out, NetCDF error',stat(i),' in writing variables, no.', i
             print *, nf_strerror(stat(i))
          endif
       END DO
    endif

    iter = iter + 1

END SUBROUTINE nc_out

subroutine nc_finalize

    USE MESH
    USE ELEMENTS
    USE SWE
    use netcdf_variables
    use mpi_communication
    
    implicit none
    
    INTEGER                    :: stat(64), istat(64), s, q, i, n, k, el           ! auxiliaries
    REAL(KIND=dp)              :: max_mwh

    CHARACTER(len=200)          :: attstr
    CHARACTER(len=3)            :: StrMeshID

! For reduced mesh (inundated area, mwh > threshold on land)        
    CHARACTER(len=100)          :: inun_ncfile
    REAL(KIND=sp), parameter    :: mwh_threshold=0.01
    INTEGER                     :: inun_nod2D, inun_elem2D
    INTEGER                     :: inun_nod(nod2D), inun_nod_inv(nod2D)
    INTEGER                     :: inun_elem_nod(3,elem2D), inun_elem(elem2D)
    REAL(KIND=dp)               :: inun_bbox_xmax, inun_bbox_xmin
    REAL(KIND=dp)               :: inun_bbox_ymax, inun_bbox_ymin    
    INTEGER                     :: iid 			     ! IDs netCDF file
    INTEGER                     :: DimId_inun_n2D			     ! dimension: nodes
    INTEGER                     :: DimId_inun_el2D			     ! dimension: elements
    INTEGER                     :: Dim1, Dim2, Dim3, Dim4	     ! dimensions: 
    INTEGER                     :: VarId_inun_trian			     ! variable: triangle
    INTEGER                     :: VarId_inun_lon, VarId_inun_lat ! variables: 2D location, longitude, latitude
    INTEGER                     :: VarId_inun_tri_area		     ! variable: area
    INTEGER                     :: VarId_inun_topo, VarId_inun_index	     ! variables: topography, index
    INTEGER                     :: VarId_inun_topo_ini                    ! variables: initial topography
    INTEGER                     :: VarId_inun_mwh  			     ! variable: maximum wave height
    INTEGER                     :: VarId_inun_flux, VarId_inun_vel             ! variables: flux and mean absolute velocity
    INTEGER                     :: VarId_inun_uplift                      ! variable: topo/bathy uplift
    INTEGER                     :: dimarray(2)
        
    real(kind=sp)               :: aux_loc1(myDim_nod2D), aux_loc2(myDim_nod2D)
    REAL(kind=sp), allocatable  :: global1(:), glob_mwh(:), glob_lat(:), glob_lon(:)
    integer, allocatable        :: globali(:), glob_elem(:,:)
    integer, allocatable        :: aux_el(:,:)

    ! Prepare a separate netcdf for the inundated area. Aim: faster post processing
    inun_ncfile = trim(OutputPath)//trim(TSID)//trim(output_prefix)//'_inundation.nc'
    inun_elem2D = 0

!------------- Preparations for the output of fields and variables   
        
    max_mwh=0.
!$OMP PARALLEL DO
    do n=1, myDim_nod2D
       max_mwh = max(max_mwh, mwh(n))
    enddo
!$OMP END PARALLEL DO

    call com_reduce_begin(max_mwh, com_max)
    
    IF (coordinate_type == 1) THEN
!$OMP PARALLEL DO       
       DO n=1,myDim_nod2D
          aux_loc1(n) = coord_nod(1,n) * rad_inv
          aux_loc2(n) = coord_nod(2,n) * rad_inv
       end DO
!$OMP END PARALLEL DO
    else
!$OMP PARALLEL DO       
       DO n=1,myDim_nod2D
          aux_loc1(n) = coord_nod(1,n)
          aux_loc2(n) = coord_nod(2,n)
       end DO
!$OMP END PARALLEL DO
    end IF
    
    call com_reduce_end

    s=0
    if (mype==0) then
       s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'globalMinSSH', NF_DOUBLE, 1, notyetknown)
       s=s+1; stat(s) = NF_PUT_ATT_DOUBLE(fid, NF_GLOBAL, 'globalMaxSSH', NF_DOUBLE, 1, real(max_mwh,dp)) 
    end if


!----------- Output of fields, both global and the inundated area
    
    if (npes==1) then
       
       s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_lon,      aux_loc1(:) ) 
       s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_lat,      aux_loc2(:) ) 
       !----- triangles
       s=s+1; stat(s) = NF_PUT_VAR_INT( fid, VarId_trian,    elem_nod(:, :) - 1) 
       s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_tri_area, REAL(voltriangle(:), sp))
       !----- nodal values    
       s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_topo_ini, nodhn_init(:)) 
       s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_topo,     REAL(nodhn(:), sp)) 
       s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_uplift,   REAL(ssh_init(:), sp)) 
       s=s+1; stat(s) = NF_PUT_VAR_INT( fid, VarId_index,    index_nod(:)) 
       s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_arrival,  arrival_time(:)) 
       s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_mwh,      mwh(:))
       s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_vel,      max_abs_vel(:))
       s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_flux,     max_flux(:))
       
    
       if (write_nc_inundation) then
          ! Reduce the mesh to the inundated area, i.e., elements with
          ! nodes on land and mwh >= threshold for at least one node
          ! in an element (we need the neighbouring nodes for the interpolation)
          
          inun_nod_inv(1:nod2D) = -1
          k=0
          DO el=1,elem2D
             if (all(nodhn_init(elem_nod(1:3,el)) <= 0) .and. &
                  any( mwh(elem_nod(1:3,el)) >= mwh_threshold) ) then
                ! collect the element
                k=k+1
                inun_elem(k) = el
                ! mark all nodes as to be collected
                inun_nod_inv(elem_nod(1:3,el)) = 0  
             endif
          ENDDO
          inun_elem2D = k
          
          if (inun_elem2D == 0) then
             print *,'!!! There is no inundated area in the computational mesh. !!!'
             print *,trim(inun_ncfile), ' will not be written.'
          else 
             
             ! Now, collect all nodes on land needed for the interpolation
             ! of the inundation
             k = 0
             DO n=1,nod2D
                if ( inun_nod_inv(n) == 0 ) then
                   k = k+1
                   inun_nod(k) = n
                   inun_nod_inv(n) = k
                endif
             END DO
             inun_nod2D = k
             
             ! Finally, adjust the numbering of the inundated elements' nodes
             DO k = 1,inun_elem2D
                inun_elem_nod(1:3,k) = inun_nod_inv(elem_nod(1:3,inun_elem(k))) -1 ! -1 for C numbering
             ENDDO
             
             call nc_inundation_preamble()
             
             !----- write variables
             q=0
             q=q+1; istat(q) = NF_PUT_VAR_INT( iid, VarId_inun_trian,   inun_elem_nod(        1:3, 1:inun_elem2D)) 
             q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_tri_area,REAL(voltriangle(inun_elem(1:inun_elem2D)),sp))
             
             q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_lon,      aux_loc1(       inun_nod(1:inun_nod2D))) 
             q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_lat,      aux_loc2(       inun_nod(1:inun_nod2D))) 
             q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_topo,     REAL(nodhn(     inun_nod(1:inun_nod2D)), sp)) 
             q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_topo_ini, REAL(nodhn_init(inun_nod(1:inun_nod2D)), sp)) 
             q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_uplift,   REAL(ssh_init(  inun_nod(1:inun_nod2D)), sp))  
             q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_mwh,      mwh(            inun_nod(1:inun_nod2D)) ) 
             
             DO i = 1, q
                IF (istat(i) /= NF_NOERR) then
                   PRINT *, 'nc_inundation, NetCDF error',istat(i),' in writing variables, no.', i
                   print *, nf_strerror(istat(i))
                end IF
             END DO
       
          endif ! inun_elem2D > 0
    
       end if ! write_nc_inundation

       
    else ! npes > 1
       ! in the parallel case, writing fields is more complicated, as we have to collect the
       ! global information first.
       ! And we want to avoid too many global fields at the same times, i.e. we have to
       ! interleave writing the global data and the data reduced to the inundated area.
       ! Further optimization is possible, partly rather straight forward (determine bounding box
       ! in parallel, allocate / deallocate aux fields like inun_nod_inv), partly tedious (setting
       ! up the inundation area in parallel and collect the bits).

       ! To determine the reduced triangulation for the inundated are, we need
       if (mype ==  0) then
          allocate(glob_elem(3,elem2D))        ! 1. triangulation
          allocate(glob_mwh(nod2D))            ! 2. mwh
          allocate(global1(max(nod2D,elem2D))) ! 3. initial topography
          allocate(glob_lat(nod2D))            ! 4. coordinates: lat
          allocate(glob_lon(nod2D))            ! 5. coordinates: lon
       else
          allocate(glob_elem(1,1))
          allocate(glob_mwh(1))
          allocate(global1(1))
          allocate(glob_lat(1))
          allocate(glob_lon(1))
       endif
       ! Would be nicer w/o ifdef, but we need myList here... 
#ifdef USE_MPI
       allocate(aux_el(3,myDim_elem2D))
       !$OMP PARALLEL DO    
       do el=1 , myDim_elem2D
          ! elem_nod is local numbering, convert to global
          ! -1 for C numbering convention used in many post processing tools
          aux_el(1:3,el) = com_nod2D%myList(elem_nod(1:3,el))  
       end do
       !$OMP END PARALLEL DO
       call com_gather(com_elem2D, aux_el, glob_elem)
       deallocate(aux_el)
#endif
       call com_gather(com_nod2D, mwh, glob_mwh)
       call com_gather(com_nod2D, nodhn_init, global1)
       call com_gather(com_nod2D, aux_loc1, glob_lon)
       call com_gather(com_nod2D, aux_loc2, glob_lat)
       
       if (mype==0) then
          if (write_nc_inundation) then
             inun_nod_inv(1:nod2D) = -1
             k=0
             DO el=1,elem2D
                if (all(global1(glob_elem(1:3,el)) <= 0) .and. &
                     any( glob_mwh(glob_elem(1:3,el)) >= mwh_threshold) ) then
                   ! collect the element
                   k=k+1
                   inun_elem(k) = el
                   ! mark all nodes as to be collected
                   inun_nod_inv(glob_elem(1:3,el)) = 0  
                endif
                glob_elem(1:3,el) = glob_elem(1:3,el)
             ENDDO
             inun_elem2D = k
             
             if (inun_elem2D == 0) then
                print *,'!!! There is no inundated area in the computational mesh. !!!'
                print *,trim(inun_ncfile), ' will not be written.'
             else 
                
                ! Now, collect all nodes on land needed for the interpolation
                ! of the inundation
                k = 0
                DO n=1,nod2D
                   if ( inun_nod_inv(n) == 0 ) then
                      k = k+1
                      inun_nod(k) = n
                      inun_nod_inv(n) = k
                   endif
                END DO
                inun_nod2D = k
                
                ! Finally, adjust the numbering of the inundated elements' nodes
                DO k = 1,inun_elem2D
                   inun_elem_nod(1:3,k) = inun_nod_inv(glob_elem(1:3,inun_elem(k))) -1  ! -1 for C numbering
                ENDDO
                
                ! And we need the bounding box here, it must be written with the other definitions
                ! and attributes, not later in the variable mode for netcdf.
                
                
                call nc_inundation_preamble()
                
                q = 0;
                q=q+1; istat(q) = NF_PUT_VAR_INT( iid, VarId_inun_trian,   inun_elem_nod(1:3, 1:inun_elem2D)) 
                q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_mwh,     glob_mwh(inun_nod( 1:inun_nod2D))) 
                q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_topo_ini,global1( inun_nod( 1:inun_nod2D)))
                q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_lon,     glob_lon(inun_nod( 1:inun_nod2D)))
                q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_lat,     glob_lat(inun_nod( 1:inun_nod2D)))
             endif  ! inundation > 0
          endif ! write_nc_inundation
          
          s = 0; 
          s=s+1; stat(s)  = NF_PUT_VAR_INT( fid, VarId_trian,     glob_elem(    1:3, 1:elem2D)-1) ! for C numbering
          s=s+1; stat(s)  = NF_PUT_VAR_REAL(fid, VarId_mwh,       glob_mwh(          1:nod2D)) 
          s=s+1; stat(s)  = NF_PUT_VAR_REAL(fid, VarId_topo_ini,  global1(1:nod2D))
          s=s+1; stat(s)  = NF_PUT_VAR_REAL(fid, VarId_lon,       glob_lon(1:nod2D))
          s=s+1; stat(s)  = NF_PUT_VAR_REAL(fid, VarId_lat,       glob_lat(1:nod2D))
          endif    ! mype==0

       
       deallocate(glob_elem, glob_mwh, glob_lon, glob_lat)
       
       call com_gather(com_nod2D, arrival_time, global1)
       if (mype==0) then
          s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_arrival,  global1)
       end if
       
       call com_gather(com_nod2D, max_abs_vel, global1)
       if (mype==0) then
          s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_vel,      global1)
       end if
       
       call com_gather(com_nod2D, max_flux, global1)
       if (mype==0) then
          s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_flux,     global1)
       endif

       call com_gather(com_nod2D, nodhn, global1)
       if (mype==0) then
          s=s+1; stat(s)  = NF_PUT_VAR_REAL(fid, VarId_topo,      global1(1:nod2D))
          if (inun_elem2D > 0) then
             q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_topo, global1(inun_nod(1:inun_nod2D)))
          endif
       end if
       call com_gather(com_nod2D, ssh_init, global1)
       if (mype==0) then
          s=s+1; stat(s)  = NF_PUT_VAR_REAL(fid, VarId_uplift,      global1(1:nod2D))
          if (inun_elem2D > 0) then
             q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_uplift, global1(inun_nod(1:inun_nod2D)))
          end if
       endif
       call com_gather(com_elem2D, voltriangle, global1)
       if (mype==0) then
          s=s+1; stat(s)  = NF_PUT_VAR_REAL(fid, VarId_tri_area,      global1(1:elem2D))
          if (inun_elem2D > 0) then
             q=q+1; istat(q) = NF_PUT_VAR_REAL(iid, VarId_inun_tri_area, global1(inun_elem(1:inun_elem2D)))
          endif
       end if
! Not (yet) available for npes > 1       
!       if (write_ttt) then
!          call com_gather(com_elem2D, ttt, global1)
!          if (mype==0) then
!             s=s+1; stat(s) = NF_PUT_VAR_REAL(fid, VarId_ttt,   global1(1:nod2D))
!          endif
!       end if
    
       deallocate(global1)
       if (mype ==  0) then
          allocate(globali(nod2D))
       else
          allocate(globali(1))
       endif
       
       call com_gather(com_nod2D, index_nod, globali)
       if (mype==0) then
          s=s+1; stat(s) =NF_PUT_VAR_INT(fid,   VarId_index,    globali(1:nod2D)) 
       end if
       
       deallocate(globali)
       
    endif
    


    if (mype==0) then
       DO i = 1, s 
          IF (stat(i) /= NF_NOERR) then
             PRINT *,  'nc_finalize, NetCDF error,',stat(i),' writing attributes/variables, no.', i
             print *, nf_strerror(stat(i))
          END IF
       END DO
       stat(1) = NF_CLOSE(fid)
       if (stat(1) /= NF_NOERR) then
          print *,  'nc_finalize, NetCDF error',stat(1),' in closing NetCDF file'
          print *, nf_strerror(stat(1))
       endif
       
       if (inun_elem2D > 0) then
          DO i = 1, q 
             IF (istat(i) /= NF_NOERR) then
                PRINT *,  'nc_inundation, NetCDF error,',istat(i),' writing attributes/variables, no.', i
                print *, nf_strerror(istat(i))
             END IF
          END DO
          istat(1) = NF_CLOSE(iid)
          IF (istat(1) /= NF_NOERR) THEN
             PRINT *, 'nc_inundation, NetCDF error',istat(1),' in closing NetCDF file'          
             print *, nf_strerror(istat(1))
          END IF
       endif
    endif
    
  contains
    
    subroutine nc_inundation_preamble

       !----- open file and write global attributes
       
       q = 0
       q=q+1; istat(q) = NF_CREATE(TRIM(inun_ncfile), 0, iid) 
       attstr  = 'Tsunami Simulation - Inundation'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'title',       LEN_TRIM(attstr), TRIM(attstr)) 
       attstr  = 'AWI'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'institution', LEN_TRIM(attstr), TRIM(attstr)) 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'Conventions',                6,     'CF-1.0') 
       
       
       WRITE(StrMeshID, '(I3.3)') MeshID
       attstr  = 'TsunAWI Revision '//revision//' - Mesh: '//StrMeshID
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'source',      LEN_TRIM(attstr), TRIM(attstr)) 
       attstr  = 'no entry'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'history',     LEN_TRIM(attstr), TRIM(attstr)) 
       attstr  = 'none'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'references',  LEN_TRIM(attstr), TRIM(attstr)) 
       attstr  = 'none'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'comment',     LEN_TRIM(attstr), TRIM(attstr)) 
       attstr  = 'AWI' 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'provider',    LEN_TRIM(attstr), TRIM(attstr)) 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'model',                      7,    'TsunAWI')
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'modelVersion',LEN_TRIM(revision),   revision)
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, NF_GLOBAL, 'scope',                     10, 'inundation')
       q=q+1; istat(q) = NF_PUT_ATT_INT( iid, NF_GLOBAL, 'degreeOfFreedom', NF_INT, 1, inun_nod2D)
              
       q=q+1; istat(q) = NF_PUT_ATT_DOUBLE(iid, NF_GLOBAL, 'integrationTime_in_seconds', NF_DOUBLE, 1, real(T_end,dp))
       
       DO i = 1,  q
          IF (istat(i) /= NF_NOERR) then
             PRINT *, 'nc_inundation, NetCDF error',istat(i),' in attribute definitions, no.', i
             print *, nf_strerror(istat(i))
          end IF
       END DO
!----- DEFINE DIMENSIONS ------------------------------

       q=0
       q=q+1; istat(q) = NF_DEF_DIM(iid, 'one',   1, dim1)                           
       q=q+1; istat(q) = NF_DEF_DIM(iid, 'two',   2, dim2)                           
       q=q+1; istat(q) = NF_DEF_DIM(iid, 'three', 3, dim3)                         
       q=q+1; istat(q) = NF_DEF_DIM(iid, 'four',  4, dim4)  
       q=q+1; istat(q) = NF_DEF_DIM(iid, 'nodes_2D',     inun_nod2D, DimId_inun_n2D)             
       q=q+1; istat(q) = NF_DEF_DIM(iid, 'elements_2D', inun_elem2D, DimId_inun_el2D)
       
       DO i = 1,  q
          IF (istat(i) /= NF_NOERR) then
             PRINT *, 'nc_inundation, NetCDF error',istat(i),' in dimension definitions, no.', i
             print *, nf_strerror(istat(i))
          end IF
       END DO
       

!----- DEFINE VARIABLES ---------------------------------
       !----- coordinates
       q = 0
       q=q+1; istat(q) = NF_DEF_VAR(iid, 'longitude',         NF_FLOAT, 1, DimId_inun_n2D, VarId_inun_lon) 
       q=q+1; istat(q) = NF_DEF_VAR(iid, 'latitude',          NF_FLOAT, 1, DimId_inun_n2D, VarId_inun_lat) 
       
       !----- grid
       dimarray = (/ dim3, DimId_inun_el2D /)
       q=q+1; istat(q) = NF_DEF_VAR(iid, 'triangles', NF_INT,   2, dimarray(1:2),      VarId_inun_trian) 
       q=q+1; istat(q) = NF_DEF_VAR(iid, 'tri_area',  NF_FLOAT, 1, DimId_inun_el2D,    VarId_inun_tri_area) 

!----- topography
       q=q+1; istat(q) = NF_DEF_VAR(iid, 'initial_topography',NF_FLOAT, 1, DimId_inun_n2D, VarId_inun_topo_ini) 
       q=q+1; istat(q) = NF_DEF_VAR(iid, 'topography',        NF_FLOAT, 1, DimId_inun_n2D, VarId_inun_topo) 
       
       !----- F I E L D S 
       q=q+1; istat(q) = NF_DEF_VAR(iid, 'max_wave_height',   NF_FLOAT, 1, DimId_inun_n2D, VarId_inun_mwh) 
       q=q+1; istat(q) = NF_DEF_VAR(iid, 'max_abs_vel',       NF_FLOAT, 1, DimId_inun_n2D, VarId_inun_vel) 
       q=q+1; istat(q) = NF_DEF_VAR(iid, 'max_flux',          NF_FLOAT, 1, DimId_inun_n2D, VarId_inun_flux) 
       q=q+1; istat(q) = NF_DEF_VAR(iid, 'initial_uplift',    NF_FLOAT, 1, DimId_inun_n2D, VarId_inun_uplift) 
       
       !----- vector variables
       
       DO i = 1,  q
          IF (istat(i) /= NF_NOERR) then
             PRINT *, 'nc_inundation, NetCDF error',istat(i),' in variable definition, no.', i
             print *, nf_strerror(istat(i))
          endif
       END DO
       
       !----- DEFINE ATTRIBUTES -----------------------------------------

       q = 0       
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_lon,      'long_name',    9, 'longitude') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_lon,      'units',       12, 'degrees_east') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_lat,      'long_name',    8, 'latitude') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_lat,      'units',       13, 'degrees_north') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_tri_area, 'long_name',   13, 'triangle area')     
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_topo_ini, 'long_name',   18, 'initial_topography') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_topo_ini, 'field',       18, 'initial_topography') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_topo_ini, 'connections', 20, 'triangles, triangles') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_topo_ini, 'positions',   17, 'surface_locations') 
       
       attstr  = 'meters (before shock bathymetry and topography)'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_topo_ini, 'units',       LEN_TRIM(attstr), attstr) 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_topo,     'long_name',   10, 'topography') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_topo,     'field',       10, 'topography') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_topo,     'connections', 20, 'triangles, triangles') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_topo,     'positions',   17, 'surface_locations') 
       
       attstr  = 'meters (after shock bathymetry and topography)'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_topo,     'units',       LEN_TRIM(attstr), attstr) 
       
       attstr='initial uplift due to earthquake'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_uplift,   'long_name',   LEN_TRIM(attstr), attstr) 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_uplift,   'field',       14, 'initial_uplift') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_uplift,   'connections', 20, 'triangles, triangles') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_uplift,   'positions',   17, 'surface_locations') 
       
       attstr  = 'maximum wave height\n (above after shock topography on land)'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_mwh,       'long_name',   LEN_TRIM(attstr), attstr) 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_mwh,       'units',        6, 'meters')
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_mwh,       'field',       15, 'max_wave_height') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_mwh,       'connections', 20, 'triangles, triangles') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_mwh,       'positions',   17, 'surface_locations')
       attstr='maximum nodal velocity approximation'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_vel,       'long_name',   LEN_TRIM(attstr), attstr) 
       
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_vel,       'units',        5, 'm/sec') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_vel,       'field',       11, 'max_abs_vel')
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_vel,       'connections', 20, 'triangles, triangles') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_vel,       'positions',   17, 'surface_locations') 
       
       attstr  = 'maximum flux in nodes\n (maximum value of the product of velocity and wave height)'
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_flux,      'long_name',   LEN_TRIM(attstr), attstr) 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_flux,     'units',         7, 'm^2/sec') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_flux,     'field',         8, 'max_flux') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_flux,     'connections',  20, 'triangles, triangles') 
       q=q+1; istat(q) = NF_PUT_ATT_TEXT(iid, VarId_inun_flux,     'positions',    17, 'surface_locations') 

       if (npes == 1) then
          inun_bbox_xmin = minval(aux_loc1(inun_nod(1:inun_nod2D)))
          inun_bbox_xmax = maxval(aux_loc1(inun_nod(1:inun_nod2D)))
          inun_bbox_ymin = minval(aux_loc2(inun_nod(1:inun_nod2D)))
          inun_bbox_ymax = maxval(aux_loc2(inun_nod(1:inun_nod2D)))
       else
          inun_bbox_xmin = minval(glob_lon(inun_nod(1:inun_nod2D)))
          inun_bbox_xmax = maxval(glob_lon(inun_nod(1:inun_nod2D)))
          inun_bbox_ymin = minval(glob_lat(inun_nod(1:inun_nod2D)))
          inun_bbox_ymax = maxval(glob_lat(inun_nod(1:inun_nod2D)))
       endif
       
       q=q+1; istat(q) = NF_PUT_ATT_DOUBLE(iid, NF_GLOBAL, 'boundingBox_UpperRight_Lon', NF_DOUBLE, 1, inun_bbox_xmax)
       q=q+1; istat(q) = NF_PUT_ATT_DOUBLE(iid, NF_GLOBAL, 'boundingBox_UpperRight_Lat', NF_DOUBLE, 1, inun_bbox_ymax)
       q=q+1; istat(q) = NF_PUT_ATT_DOUBLE(iid, NF_GLOBAL, 'boundingBox_LowerLeft_Lon',  NF_DOUBLE, 1, inun_bbox_xmin)
       q=q+1; istat(q) = NF_PUT_ATT_DOUBLE(iid, NF_GLOBAL, 'boundingBox_LowerLeft_Lat',  NF_DOUBLE, 1, inun_bbox_ymin)

       DO i = 1, q
          IF (istat(i) /= NF_NOERR) then
             PRINT *, 'nc_inundation, NetCDF error',istat(i),' in attribute assignments, no.', i
             print *, nf_strerror(istat(i))
          end IF
       END DO

       istat(1) = NF_ENDDEF(iid)
       if (istat(1) /= NF_NOERR) then
          PRINT *, 'nc_inundation, NetCDF error',istat(i),' in nf_enddef'
          print *, nf_strerror(istat(i))
       end IF
    end subroutine nc_inundation_preamble
    
    
  END SUBROUTINE nc_finalize
#endif
